"""Module to show how a tool can vbe created from ms thermo.

Here we create a simple CLI for the Gasout tool"""

import os


"""Example of gasout use.

Here this is a tool for the CFL LES solver AVBP
"""
import yaml
import copy
import h5py
import numpy as np
import hdfdict
from ms_thermo.gasout import gasout_with_input, gasout_dump_default_input
from ms_thermo import State


def load_mesh_and_solution(fname, mname):
    """Read HDF5 mesh and solution files"""
    print("Reading solution in ", fname)
    with h5py.File(fname, "r") as fin:
        sol = hdfdict.load(fin, lazy=False)
    print("Reading mesh in ", mname)
    with h5py.File(mname, "r") as fin:
        mesh = hdfdict.load(fin, lazy=False)
    return mesh, sol


def build_data_from_avbp(mesh, sol):
    """Buid the data set from"""
    state = State.from_cons(
        sol["GaseousPhase"]["rho"], sol["GaseousPhase"]["rhoE"], sol["RhoSpecies"]
    )

    x_coor = mesh["Coordinates"]["x"]
    y_coor = mesh["Coordinates"]["y"]
    try:
        z_coor = mesh["Coordinates"]["z"]
        coor = np.stack((x_coor, y_coor, z_coor), axis=-1)
    except KeyError:
        coor = np.stack((x_coor, y_coor), axis=-1)

    return coor, state


def save_data_for_avbp(state, sol, fname):
    """Update the full solution with the state parameters"""
    sol_new = copy.deepcopy(sol)
    sol_new["GaseousPhase"]["rho"] = state.rho
    sol_new["GaseousPhase"]["rhoE"] = state.rho * state.energy
    for spec in sol_new["RhoSpecies"]:
        sol_new["RhoSpecies"][spec] = state.rho * state.mass_fracs[spec]

    try:
        sol_new["Additionals"]["temperature"] = state.temperature
        sol_new["Additionals"]["pressure"] = state.pressure
    except KeyError:
        print("-Additionals- group is not part of the solution.")

    print("Saving solution in ", fname)
    with h5py.File(fname, "w") as fout:
        hdfdict.dump(sol_new, fout)


def gasout_tool(inputfile):
    """Main call"""

    with open(inputfile, "r") as fin:
        in_nob = yaml.load(fin, Loader=yaml.Loader)

    mesh, sol = load_mesh_and_solution(in_nob["inst_solut"], in_nob["mesh"])

    coor, state = build_data_from_avbp(mesh, sol)

    state_new, log = gasout_with_input(coor, state, in_nob)

    print(log)

    save_data_for_avbp(state_new, sol, in_nob["inst_solut_output"])



"""Module with functions helpers to  create gasout for CFD solvers"""

import numpy as np


HEADLOG = """
   ### Gasout tool from ms_thermo package ###

    This is a demonstrator! 
    It has not been thoroughy tested yet!

"""

__all__ = [
    "gasout_with_input",
    "gasout_dump_default_input",
    "alter_state",
    "spherical_tanh_mask",
    "directional_linear_mask",
    "fraction_z_mask",
]


def alter_state(state, alpha, temp_new=None, press_new=None, y_new=None, verbose=False):
    """ Alter the state field rho;rhoE, rhoY

    :param state: a ms_thermo State object od size n points.
    :param alpha: a mask of alteration shape(n)
        btw 0 (no alteration) to 1 (full alteration)
    
    :param press_new: (optionnal) the new mass pressure to apply
         either float or numpy array of shape (n)
    :param temp_new: (optionnal) the new temperature to apply
        either float or numpy array of shape (n)
    :param y_new: (optionnal) the new mass fractions to apply
        dict of species {"O2": 0.1, "N2", 0.9}
        values either floats or numpy arrays of shape (n)
    """
    log = str()
    if temp_new is not None:
        temp = np.array(state.temperature)
        state.temperature = temp + alpha * (temp_new - temp)

    if y_new is not None:
        y_mix = dict()
        state_species = list(state.mass_fracs.keys())

        for spec in y_new:
            if spec not in state_species:
                msgerr = "\n\n Species mismatch:"
                msgerr += "\n" + str(spec) + " is not part of the mixture:\n"
                msgerr += "/".join(state_species)
                msgerr += "\nCheck your input file pretty please..."
                raise RuntimeWarning(msgerr)

        for spec in state.mass_fracs:
            value = 0
            if spec in y_new:
                value = y_new[spec]

            y_mix[spec] = state.mass_fracs[spec] + alpha * (
                value - state.mass_fracs[spec]
            )
        state.mass_fracs = y_mix

    if press_new is not None:
        state.pressure = state.pressure + alpha * (press_new - state.pressure)

    if verbose:
        footprint = 100.0 * np.sum(alpha) / alpha.shape[0]
        log += f"Footprint : {footprint: .3f} %"
    return log


def spherical_tanh_mask(coor, center=None, radius=None, delta=None, verbose=False):
    """Compute a spherical mask.

    :param coor: the n points coordinates
        a numpy array of shape (n, ndim)

    :param center: center coordinates of gasout [m]
            a list/tuple/nparray of size ndim
    :param radius: the radius of gasout [m]
    :param delta: the transition of gasout [m]

    :return:
        alpha, a  mask of floats shape (n)
        0 if no mask
        1 if masked

    """
    if center is None:
        center = np.average(coor, axis=0)
    if radius is None:
        size = min(np.max(coor, axis=0) - np.min(coor, axis=0))
        radius = 0.2 * size
    if delta is None:
        delta = 0.5 * rad_0

    center = np.array(center)
    center_ndim = center.shape[0]
    coor_ndim = coor.shape[-1]
    if center_ndim != coor_ndim:
        msg_err = "\n\nDimension mismatch in the spherical center"
        msg_err += "\nCenter is " + str(center_ndim) + "D"
        msg_err += "\nCoords are " + str(coor_ndim) + "D"
        msg_err += "\nCheck your input file pretty please..."
        raise RuntimeWarning(msg_err)

    r_coor = np.linalg.norm(coor - center, axis=1)
    alpha = 0.5 * (1.0 - np.tanh((r_coor - radius) / (delta / 2)))
    return alpha


def directional_linear_mask(
    coor, axis=0, startstop=None,
):
    """Compute a linear transition mask

    :param coor: the n points coordinates
        a numpy array of shape (n, ndim)
    :param axis: 0 x, 1 y , 2 z
    :param startstop: start and stop on the direction [m]

    :return:
        alpha, a  mask of floats shape (n)
        0 if no mask
        1 if masked

    """
    direction = coor[:, axis]
    center = np.average(direction)
    size = np.max(direction) - np.min(direction)
    if startstop is None:
        startstop = (center - 0.2 * size, center + 0.2 * size)

    alpha = (direction - startstop[0]) / (startstop[1] - startstop[0])
    alpha = np.clip(alpha, 0.0, 1.0)

    return alpha


def fraction_z_mask(
    state,
    specfuel,
    zmin=0.3,
    zmax=0.7,
    fuel_mass_fracs=None,
    oxyd_mass_fracs=None,
    atom_ref=None,
    verbose=False,
):
    """Compute a mask found by a transition in Z farction

    :param state: a ms_thermo state object
    :specfuel: string the fuel species name
    :param z_min: mask disabled (0) below this value
    :param z_max: mask disabled (0) over this value

    :return:
        alpha, a  mask of floats shape (n)
        0 if no mask
        1 if masked

    """
    z_frac = state.compute_z_frac(
        specfuel,
        fuel_mass_fracs=None,
        oxyd_mass_fracs=None,
        atom_ref="C",
        verbose=False,
    )

    z_mid = 0.5 * (zmax + zmin)
    z_gap = 0.5 * (zmax - zmin)
    alpha = np.where(abs(z_frac - z_mid) < z_gap, 1.0, 0)
    return alpha


def gasout_with_input(coor, state, in_nob):
    """Gasout function based upon a nest obeject input

    :param coor: the n points coordinates
        a numpy array of shape (n, ndim)

    :param state: a ms_thermo State object od size n points.
    :param in_nob: a nested object for input,
        see gasout_dump_default_input fro reference

    """

    log = str(HEADLOG)
    log += "\nInitial state"
    log += "\n ==========="
    log += "\n" + state.__repr__()

    for i, action in enumerate(in_nob["actions"]):

        log += "\n\nAction " + str(i) + ":" + action["type"]

        if action["type"] == "directional_linear_mask":
            reqd_keys = [
                "direction",
                "max_abcissa",
                "min_abcissa",
                "new_pressure",
                "new_temperature",
                "new_yk",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            if action["direction"] == "x":
                axis = 0
            elif action["direction"] == "y":
                axis = 1
            elif action["direction"] == "z":
                axis = 2
            else:
                msgerr = "\n INPUT ERROR : " + "Direction must be one of x, y or z"
                raise RuntimeWarning(log + msgerr)

            alpha = directional_linear_mask(
                coor,
                axis=axis,
                startstop=[action["min_abcissa"], action["max_abcissa"]],
            )

        elif action["type"] == "spherical_tanh_mask":
            reqd_keys = [
                "center",
                "radius",
                "delta",
                "new_pressure",
                "new_temperature",
                "new_yk",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = spherical_tanh_mask(
                coor,
                center=action["center"],
                radius=action["radius"],
                delta=action["delta"],
            )

        elif action["type"] == "fraction_z_mask":
            reqd_keys = [
                "specfuel",
                "atom_ref",
                "oxyd_mass_fracs",
                "fuel_mass_fracs",
                "zmax",
                "zmin",
                "new_pressure",
                "new_temperature",
                "new_yk",
            ]

            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = fraction_z_mask(
                state,
                action["specfuel"],
                zmin=action["zmin"],
                zmax=action["zmax"],
                fuel_mass_fracs=action["fuel_mass_fracs"],
                oxyd_mass_fracs=action["oxyd_mass_fracs"],
                atom_ref=action["specfuel"],
                verbose=True,
            )

        out = alter_state(
            state,
            alpha,
            temp_new=action["new_temperature"],
            press_new=action["new_pressure"],
            y_new=action["new_yk"],
            verbose=True,
        )
        log += "\n" + out

    log += "\n\n Final state"
    log += "\n ==========="

    log += "\n" + state.__repr__()

    return state, log


def gasout_dump_default_input(fname="gasout_in.yml"):
    """Dump a Yamlf file adapted to control gasout_with_input

    Note: i Stored this with a string to include
    the comments in the YAML. Not my best code...
    """

    default_input = """
#This is the tentative input file of a pytohon Gasout.
# Enable YAML syntax highlighting in you IDE for the best experience.

# input files
inst_solut: ./solut_0003.sol.h5
mesh: ./solut_0003.mesh.h5
inst_solut_output: ./gasouted.sol.h5

# list (each - block) of actions to take
actions:
- type: directional_linear_mask
  direction: "x"             #direction x, y, or z
  max_abcissa: 0.11        #If None, takes  70% of CAO size
  min_abcissa: 0.1         #If None, takes  30% of CAO size
  new_pressure: null       #use null if you dont want to edit this field
  new_temperature: 2000.0  # Temp in K.
  new_yk: null

- type: spherical_tanh_mask
  center: [0.1, 0.1, 0]    # center sphere. If None, takes CAO center.
  radius: 0.01              # radius sphere [m]. If None, takes from CAO size
  delta: 0.05              # transition thickness [m], If None, takes half rad
  new_pressure: null
  new_temperature: 1600.
  new_yk:                  # dictionnary of species Yk
    N2: 0.8                # order and competion does not matter
    O2: 0.2                # Sum MUST be 1!

- type: fraction_z_mask
  specfuel: KERO_LUCHE     # Fuel spec name
  atom_ref: C              # reference atom for Z computation
  oxyd_mass_fracs:                 # Mixture oxidizer, for Z computation.
    O2: 0.233              #   If None, takes AIR
    N2: 0.767
  fuel_mass_fracs:  None          # Mixture fuel. dict of Yk.
  zmax: 0.02              #   If None, takes the mix at the Higesh Yk fuel
  zmin: 0.01
  new_pressure: 103000.0  # Pressure in Pa
  new_temperature: null
  new_yk: null

- type: fraction_z_mask    # You can repeat several treatments
  specfuel: KERO_LUCHE
  atom_ref: C
  oxyd_mass_fracs:
    O2: 0.233
    N2: 0.767
  fuel_mass_fracs:  None
  zmax: 0.05
  zmin: 0.04
  new_pressure: 103000.0
  new_temperature: null
  new_yk: null
"""
    print("A default input has been created: " + fname)
    with open(fname, "w") as fin:
        fin.write(default_input)


def _check_action_params(reqd_keys, provided_keys):
    """Feedback on the keys provided"""
    log = str()
    for key in provided_keys:
        if key not in reqd_keys + ["type"]:
            log += "\n.   - key " + key + " not accepted"
    if log != "":
        log += "\n.  - Accepted keys"
        log += "\n.      >".join(reqd_keys)

    for key in reqd_keys:
        if key not in provided_keys:
            log += "\n key " + key + " is missing."
    return log



#!/usr/bin/env python
"""
cli.py

Command line interface for tools in ms_thermo
"""


import click
from pkg_resources import resource_listdir, resource_string
from ms_thermo.tadia import tadia_table, tadia_cantera
from ms_thermo.yk_from_phi import yk_from_phi
from ms_thermo.fresh_gas import fresh_gas as fg_
from ms_thermo.example_gasout import gasout_tool
from ms_thermo.gasout import gasout_dump_default_input


DPCTD = "Deprecated command from python package ms-thermo..."

# OLD COMMANDS----------------------
def redirect_fresh_gas():
    """redicrection of former command"""
    print(DPCTD)
    print(" ? Did you mean : > ms_thermo fresh-gas")


def redirect_tadia_table():
    """redicrection of former command"""
    print(DPCTD)
    print(" ? Did you mean : > ms_thermo tadia")


def redirect_tadia_cantera():
    """redicrection of former command"""
    print(DPCTD)
    print(" ? Did you mean : > ms_thermo hp-equil")


def redirect_yk_from_phi():
    """redicrection of former command"""
    print(DPCTD)
    print(" ? Did you mean : > ms_thermo yk_from_phi")


def redirect_gasout():
    """redicrection of former command"""
    print(DPCTD)
    print(" ? Did you mean : > ms_thermo gasout")


# --------------------------------------


@click.group()
def main_cli():
    """---------------    MS_THERMO  --------------------

You are now using the Command line interface of MS-Thermo,
a Python3 helper for reactive multispecies computation, created at CERFACS (https://cerfacs.fr).

This is a python package currently installed in your python environement.
See the full documentation at : https://ms-thermo.readthedocs.io/en/latest/.
"""
    pass


@click.command()
@click.argument("temperature", nargs=1)
@click.argument("pressure", nargs=1)
@click.argument("phi", nargs=1)
def tadia(temperature, pressure, phi):
    """adiabatic flame temperature for Kerosene.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry. 

    The computation is done by interpolation in a precomputed 2S_KERO_BFER table.
    The system considered to build the table was a 1D flame at constant pressure.
    """
    burnt_temperature, yk_ = tadia_table(float(temperature), float(pressure), float(phi))
    print(
        "\nThe adiabatic flame temperature of a mix "
        + f"C10H22-air from tables is : {burnt_temperature:.2f} K."
    )
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    for specie in yk_:
        print(f"{specie:12s}|       {yk_[specie]:.3f}")


main_cli.add_command(tadia)


@click.command()
@click.argument("temperature", nargs=1, type=float)
@click.argument("pressure", nargs=1, type=float)
@click.argument("phi", nargs=1, type=float)
def fresh_gas(temperature, pressure, phi):
    """description of a non-reacted kero-air mixture.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry. 

    The computation is done using a 100K piecewise linear table of enthalpies.
    """
    rho, rhoE, rhoyk = fg_(temperature, pressure, phi)
    print(f"\nrho       |  {rho:.3f} kg/m3")
    print(f"rhoE      |  {rhoE:.3f} J.kg/m3")
    print("rhoYk     |")
    for specie in rhoyk:
        print(f" {specie:9s}|  {rhoyk[specie]:.3f} mol.kg/m3")
    print("------------------------------")
    print("Yk        |")
    for specie in rhoyk:
        print(f" {specie:9s}|  {rhoyk[specie]/rho:.3f} [-]")


main_cli.add_command(fresh_gas)


@click.command()
@click.argument("temperature", nargs=1, type=float)
@click.argument("pressure", nargs=1, type=float)
@click.argument("phi", nargs=1, type=float)
@click.argument("fuel", nargs=1)
@click.argument("path2cti", nargs=1)
def hp_equil(temperature, pressure, phi, fuel, path2cti):
    """HP equilibrium using Cantera.

    The initial TEMPERATURE is in Kelvins.
    The initial PRESURE in in Pascals.
    The PHI equivalence ratio equals 1 at stoechiometry. 
    The FUEL is the name of the fuel in your system.
    The path2cti is the path to your .CTI file. 

    The computation is done using the Cantera library 
    installed on your environement (not included in this package)
    The system considered is an HP equilibrium.
    
    See http://www.cerfacs.fr/cantera/ for more information and some .CTI files

    """
    burnt_temperature, yk_ = tadia_cantera(
        float(temperature), float(pressure), float(phi), str(fuel), str(path2cti)
    )
    print(
        "\nThe adiabatic flame temperature of a mix "
        + f"{tr(fuel)}-air from cantera is : {burnt_temperature:.2f} K."
    )
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    nb_other_species = 0
    y_other_species = 0
    for specie in yk_:
        if yk_[specie] < 1e-3:
            nb_other_species += 1
            y_other_species += yk_[specie]
        else:
            print(f"{specie:12s}|       {yk_[specie]:.3f}")
    print(f"+ {nb_other_species} others |       {y_other_species:.3f}")


main_cli.add_command(hp_equil)


@click.command()
@click.argument("phi", nargs=1, type=float)
@click.argument("c_x", nargs=1, type=float)
@click.argument("h_y", nargs=1, type=float)
def yk_from_phi(phi, c_x, h_y):
    """Mass fractions of a fuel-air mixture.
    
    The PHI equivalence ratio equals 1 at stoechiometry. 
    C_X is the number of carbon atoms in your fuel.
    H_Y is the number of hydrogen atoms in your fuel
    """
    yk_ = yk_from_phi(float(c_x), float(h_y), float(phi))
    print("\nSpecies     |    Mass fraction")
    print("------------------------------")
    for specie in yk_:
        print(f"{specie:12s}|       {yk_[specie]:.3f}")


main_cli.add_command(yk_from_phi)


@click.command()
@click.option("--new", is_flag=True, help="prefill INPUTFILE with default values")
@click.argument("inputfile", nargs=1)
def gasout(inputfile, new):
    """Apply GASOUT actions to a CFD field.
    
    A GASOUT action is a local alteration of the thermodynamical
    state (temperature, pressure or mixture.)of a CFD field.
    
    The INPUTFILE, in .yaml format, controls the list of actions.
    Create a new INPUTFILE with the --new flag to see all possibilities. 
    """
    if new:
        gasout_dump_default_input(inputfile)
    else:
        gasout_tool(inputfile)


main_cli.add_command(gasout)



"""Module to build thermodynamic properties of species."""
import numpy as np
from scipy.interpolate import interp1d
from scipy.misc import derivative
from ms_thermo.constants import GAS_CST, ATOMIC_WEIGHTS

__all__ = [
    "build_thermo_from_avbp",
    "build_thermo_from_cantera",
    "build_thermo_from_chemkin",
]


class _NasaPolys:
    r"""
    Container class building NASA 7-coeffs polynomials \
    to build functions evaluating species thermodynamic properties per \
    mass unit. mass enthalpy of species k for instance is computed as : \

    .. math:: \frac{R}{W_k} (a_{5,k} + sum{i=0}^4 \frac{a_{i,k}}{i+1}T^{i+1})
    """

    def __init__(self, coeffs, t_range, molecular_weight):
        """ Instanciation  of _NasaPolys

            Parameters :
            ==========
            coeffs : a dictionary of NASA 7 polynomials corresponding
                     to "low" and "high" temperature ranges
            t_range : a dictionary of three entries: min, mid and max
                      separating temperature low range ([T_min, T_mid]) and
                      high range ([T_mid, T_max])
            molecular_weight : species molecular weight
        """

        self.enthalpy_polys = dict()
        self.enthalpy_polys["high"] = None
        self.enthalpy_polys["low"] = None
        self.t_range = t_range
        self.molecular_weight = molecular_weight
        for k in ["high", "low"]:
            # Build enthalpy polynomials
            # ie. h_k(T) = a_5 + \sum_{i=0}^4 a_{i,k}/(i+1)T^{i+1}
            poly = np.zeros(6)
            poly[0] = coeffs[k][5]
            for i in range(5):
                poly[i + 1] = coeffs[k][i] / (i + 1)
            # transform polynomial coeffs to mass and SI unit(J/kg)
            poly *= GAS_CST / molecular_weight
            self.enthalpy_polys[k] = np.polynomial.polynomial.Polynomial(poly)

    def get_total_enthalpy(self, temp):
        r"""Computing species total enthalpy as :
            $\frac{R}{W_k} (a_{5,k} + sum{i=0}^4 \frac{a_{i,k}}{i+1}T^{i+1})$
            depending on the range of temperature i.e: "low" range polynomial
            is used if temperature is lower than T_mid, and "high" range is
            used otherwise

            Parameters :
            ==========
            temp : a temperature array for which enthalpy is evaluated

            Returns :
            =======
            enthalpy : species total enthalpy array
        """
        _temp = np.float_(temp)
        condlist = [
            _temp < self.t_range["mid"],
            np.logical_and(_temp >= self.t_range["mid"], _temp <= self.t_range["max"]),
        ]
        funclist = [self.enthalpy_polys["low"], self.enthalpy_polys["high"]]
        enthalpy = np.piecewise(_temp, condlist, funclist)
        enthalpy -= self.enthalpy_polys["low"](0)
        return enthalpy

    # def get_formation_enthalpy(self):
    #     """Computing standard enthalpy of formation"""
    #     t_0 = 298.15
    #     return self.get_total_enthalpy(t_0)

    def get_total_energy(self, temp):
        r"""Computing total mass energy as : $e_k(T) = h_k(T) -RT=W_k$

            Parameters :
            ==========
            temp : a temperature array for which energy is evaluated

            Returns :
            =======
            energy : species total energy array
        """

        enthalpy = self.get_total_enthalpy(temp)
        energy = enthalpy - GAS_CST * temp / self.molecular_weight
        return energy


def _get_mol_weight_from_elements(elements):
    r""" Computes species molecular weight given atomic composition as :
         $W_k = \sum_{i=1}^N_{el} b_{i,k} w_i$

         Parameters :
         ==========
         element : a dictionary mapping species elements names to their numbers

         Returns :
         =======
         molecular_weight : species molecular weight
    """
    molecular_weight = 0.0
    for k in elements:
        molecular_weight += ATOMIC_WEIGHTS[k.lower()] * elements[k]
    return molecular_weight


def _get_species_indexes(database_file, db_format):
    """ Get indexes of species entries in a thermo database file

        Parameters :
        ==========
        database_file : full path to database file
        db_format : the database format, possible values :
                    - cantera / - chemkin / -avbp
        Returns :
        =======
        db_stream : the stream (content) of the database file
        indexes : list of species line indexes in the database file
    """
    with open(database_file, "r") as fin:
        db_stream = fin.readlines()

    indexes = []

    for i, line in enumerate(db_stream, 0):
        line = line.strip()
        cond = any(
            [
                (db_format == "avbp") and (line.startswith("species_name =")),
                (db_format == "chemkin") and (line.endswith(" 1")),
                (db_format == "cantera") and (line.startswith("species(")),
            ]
        )
        if cond:
            indexes.append(i)
    return db_stream, indexes


def _make_props_avbp(species_description):
    """ Get species properties from AVBP like thermo database

        Parameters :
        ==========
        species_description : a list of strings containing species information

        Returns :
        =======
        props : a dictionary holding species properties. It contains the keys:
                - species_name : species name
                - molecular_weight : species molecular weight
                - formation_enthalpy : species standard enthalpy of formation
                - chemical_energy : a temperature function returning species
                                    total energy
                - total_enthalpy : a temperature function returning species
                                    total enthalpy
    """

    def _avbp_str_to_float(str_val):
        """
        *Convert the given string to a float.*

        :param str_val: A string type value
        :returns: A float type value
        """
        number = str_val
        for sub in ["d", "D"]:
            number = number.replace(sub, "e")
        number = float(number)
        return number

    props = dict()
    props["species_name"] = [
        line.split()[-1]
        for line in species_description
        if line.startswith("species_name =")
    ][0]

    props["atoms"] = dict()

    for line in species_description:
        if line.startswith("species_molecular_weight ="):
            number = line.strip().split("=")[-1]
            props["molecular_weight"] = _avbp_str_to_float(number)
        # if line.startswith("species_formation_enthalpy ="):
        #     number = line.strip().split('=')[-1]
        #     props["formation_enthalpy"] = avbp_str_to_float(number)

        if line.startswith("species_C_atoms"):
            number = line.strip().split("=")[-1]
            props["atoms"]["C"] = _avbp_str_to_float(number)
        if line.startswith("species_H_atoms"):
            number = line.strip().split("=")[-1]
            props["atoms"]["H"] = _avbp_str_to_float(number)
        if line.startswith("species_O_atoms"):
            number = line.strip().split("=")[-1]
            props["atoms"]["O"] = _avbp_str_to_float(number)
        if line.startswith("species_N_atoms"):
            number = line.strip().split("=")[-1]
            props["atoms"]["N"] = _avbp_str_to_float(number)

        if line.startswith("species_enthalpy_table ="):
            # Reading and computing table of molar sensible enthalpy
            # i.e : $h_s = int_{T_0}^{T} C_{p,k}^m dT$
            # given for the temperature range [0, 5000] whose step is 100K
            number = line.strip().split("=")[-1]

            number = np.array([_avbp_str_to_float(k) for k in number.split() if k])
            number -= number[0]

            t_range = np.linspace(0.0, 5000.0, 51)
            e_sens_chem_tbl = np.zeros_like(t_range)
            total_enth_tbl = np.zeros_like(t_range)

            if len(t_range) == len(number):
                # Computing the sum of formation and sensible MOLAR enthalpy
                # i.e $h_t = h_s + \Delta h^{0,m}_{f,k}$
                total_enth_tbl = number  # + props["formation_enthalpy"]
                # Computing the molar chemical energy
                # i.e $e_k^m = h_t - RT $
                e_sens_chem_tbl = total_enth_tbl - GAS_CST * t_range

    # Building the function returning the sum of chemical
    # and sensible MASS energy given the temperature
    if props["molecular_weight"] > 0:
        e_sens_chem_tbl /= props["molecular_weight"]
        total_enth_tbl /= props["molecular_weight"]
        # props["formation_enthalpy"] /= props["molecular_weight"]

    props["chemical_energy"] = interp1d(t_range, e_sens_chem_tbl, kind="linear")
    props["total_enthalpy"] = interp1d(t_range, total_enth_tbl, kind="linear")
    return props


def _make_props_chemkin(species_description):
    """ Get species properties from Chemkin like thermo database

        Parameters :
        ==========
        species_description : a list of strings containing species information

        Returns :
        =======
        props : a dictionary holding species properties. It contains the keys:
                - species_name : species name
                - molecular_weight : species molecular weight
                - formation_enthalpy : species standard enthalpy of formation
                - chemical_energy : a temperature function returning species
                                    total energy
                - total_enthalpy : a temperature function returning species
                                    total enthalpy
    """
    props = dict()
    props["species_name"] = species_description[0][0:18].strip()

    line = species_description[0].strip()
    t_range = dict()
    elements = dict()
    coeffs = dict()

    coeffs["high"] = np.zeros(7)
    coeffs["low"] = np.zeros(7)

    t_range["min"] = float(line[45:55])
    t_range["max"] = float(line[55:65])
    t_range["mid"] = float(line[65:73])

    for i in range(4):
        elem_list = line[24 + i * 5 : 29 + i * 5].strip().split()
        if elem_list:
            elements[elem_list[0]] = int(elem_list[1])

    line = species_description[1].split("\n")[0]
    for i in range(5):
        coeffs["high"][i] = float(line[i * 15 : (i + 1) * 15])

    line = species_description[2].split("\n")[0]
    for i in range(2):
        coeffs["high"][i + 5] = float(line[i * 15 : (i + 1) * 15])
    for i in range(2, 5):
        coeffs["low"][i - 2] = float(line[i * 15 : (i + 1) * 15])

    line = species_description[3].split("\n")[0]
    for i in range(4):
        coeffs["low"][i + 3] = float(line[i * 15 : (i + 1) * 15])

    # computing molecular weight
    props["molecular_weight"] = _get_mol_weight_from_elements(elements)

    nasa = _NasaPolys(coeffs, t_range, props["molecular_weight"])
    # props["formation_enthalpy"] = nasa.get_formation_enthalpy()
    props["chemical_energy"] = nasa.get_total_energy
    props["total_enthalpy"] = nasa.get_total_enthalpy
    props["atoms"] = elements
    return props


def _make_props_cantera(species_description):
    """ Get species properties from Cantera like thermo database

        Parameters :
        ==========
        species_description : a list of strings containing species information

        Returns :
        =======
        props : a dictionary holding species properties. It contains the keys:
                - species_name : species name
                - molecular_weight : species molecular weight
                - formation_enthalpy : species standard enthalpy of formation
                - chemical_energy : a temperature function returning species
                                    total energy
                - total_enthalpy : a temperature function returning species
                                    total enthalpy
    """
    props = dict()
    props["species_name"] = [
        line for line in species_description if line.startswith("species(name")
    ][0].strip()
    for sub in ['"', "'", ",", "species(name", "="]:
        props["species_name"] = props["species_name"].replace(sub, "")

    elements = dict()
    t_range = dict()
    coeffs = dict()

    line = ""
    for line in species_description:
        if all([k in line for k in ("=", ":", ",", "atoms")]):
            break
    line = " ".join(line.split())
    for k in ("=", ",", "atoms", "'", '"'):
        line = line.replace(k, "")

    for k in line.split(" "):
        elements[k.split(":")[0]] = float(k.split(":")[1])
    props["molecular_weight"] = _get_mol_weight_from_elements(elements)

    line = [
        k
        for k in "".join(
            [" ".join(k.strip().split()) for k in species_description]
        ).split("=")
        if "NASA" in k
    ][0]
    line = line.split("NASA")

    for k in line:
        k = [j for j in k.split("(") if j]
        if k:
            k = k[0].split(")")[0]
            k = " ".join([j for j in k.split("]") if j])
            k = [j.replace(",", " ").strip() for j in k.split("[") if j]
            if len(k) == 2:
                t_1 = float(k[0].split()[0])
                t_2 = float(k[0].split()[1])
                if t_1 >= 1000.0:
                    zone = "high"
                    t_range["mid"] = t_1
                    t_range["max"] = t_2
                else:
                    zone = "low"
                    t_range["min"] = t_1
                    t_range["mid"] = t_2
                coeffs[zone] = np.array([j for j in k[1].split() if j], dtype=float)

    nasa = _NasaPolys(coeffs, t_range, props["molecular_weight"])
    # props["formation_enthalpy"] = nasa.get_formation_enthalpy()
    props["chemical_energy"] = nasa.get_total_energy
    props["total_enthalpy"] = nasa.get_total_enthalpy
    props["atoms"] = elements
    return props


class Species:
    """
    Class holding species properties
    """

    # def __init__(self, mol_weight, form_enthalpy, enthalpy_func, energy_func):
    def __init__(self, mol_weight, enthalpy_func, energy_func, atoms=None):

        """
        Instanciation of Species object

            Parameters :
            ==========
            species_name : species name
            molecular_weight : species molecular weight
            formation_enthalpy : species standard enthalpy of formation
            chemical_energy : a temperature function returning species
                                total energy
            total_enthalpy : a temperature function returning species
                                total enthalpy
        """
        self.total_enthalpy_func = enthalpy_func
        self.chemical_energy_func = energy_func
        self.molecular_weight = mol_weight
        self.atoms = atoms
        # self.formation_enthalpy = form_enthalpy

    def chemical_energy(self, temp):
        """
        Computing species total energy (sum of sensible and formation)

        :param temperature: Temperature at which energy is evaluated

        :returns: **energy** - Array of species total energy at temperature *temp*
        """
        return self.chemical_energy_func(temp)

    def total_enthalpy(self, temp):
        """
        *Computing species total enthalpy (sum of sensible and formation)*

        :param temperature: Temperature at which enthalpy is evaluated
        :type temperature: float

        :returns: **enthalpy** - Array of species total enthalpy at temperature *temp*

        """
       
        return self.total_enthalpy_func(temp)
    
    def c_p(self, temp):
        """
        Computing Specific heat at constant pressure

        :param temperature: Temperature at which C_p is evaluated

        :returns: **C_p** - Array of C_p at temperature *temp* in [J/(kg K)]
        """
        c_p = derivative(self.total_enthalpy_func, temp, dx=1.0 )
        
        return c_p
    
    def c_v(self, temp):
        """
        Computing Specific heat at constant volume

        :param temperature: Temperature at which C_v is evaluated

        :returns: **C_v** - Array of C_v at temperature *temp* in [J/(kg K)]
        """
        c_v = derivative(self.chemical_energy_func, temp, dx=1.0 )
        return c_v

    def gamma(self, temp):
        """
        Computing Specific specific heat ratio

        :param temperature: Temperature at which gamma is evaluated

        :returns: **gamma** - Array of gamma at temperature *temp*
        """
        gamma = self.c_p(temp)/self.c_v(temp)
        
        return gamma


def build_thermo_from_avbp(database_file):
    """
    *Reading all AVBP database species and storing in a dict( ) \
    whose keys are species names*

    :param database_file: Full path to database file
    :type database_file: str

    :returns: **species** -  A dict( ) of Species objects whose keys are species names
    """
    species = dict()
    _db_stream, indexes = _get_species_indexes(database_file, "avbp")

    for i, _ in enumerate(indexes, 0):
        start = indexes[i]
        if i < len(indexes) - 1:
            end = indexes[i + 1]
        else:
            end = len(_db_stream)

        species_description = [line for line in _db_stream[start:end] if " = " in line]
        props = _make_props_avbp(species_description)
        species[props["species_name"]] = Species(
            props["molecular_weight"],
            # props['formation_enthalpy'],
            props["total_enthalpy"],
            props["chemical_energy"],
            atoms=props["atoms"],
        )
    return species


def build_thermo_from_chemkin(database_file):
    """
    *Reading all* **CHEMKIN** *database species and storing in a dict( ) \
    whose keys are species names.*

    :param database_file: Full path to database file
    :type database_file: str

    :returns: **species** - A dict( ) of Species objects whose keys are species names.
    """
    species = dict()
    _db_stream, indexes = _get_species_indexes(database_file, "chemkin")

    for i in indexes:
        species_description = _db_stream[i : i + 4]
        props = _make_props_chemkin(species_description)
        species[props["species_name"]] = Species(
            props["molecular_weight"],
            # props['formation_enthalpy'],
            props["total_enthalpy"],
            props["chemical_energy"],
            atoms=props["atoms"],
        )
    return species


def build_thermo_from_cantera(database_file):
    """
    *Reading all* **CANTERA** *database species and storing in a dict( ) \
    whose keys are species names*

    :param database_file: Full path to database file
    :type database_file: str

    :returns: **species** - A dict( ) of Species objects whose keys are species names
    """

    species = dict()
    _db_stream, indexes = _get_species_indexes(database_file, "cantera")

    n_species = len(indexes)
    indexes.append(len(_db_stream))

    for i in range(n_species):
        species_description = _db_stream[indexes[i] : indexes[i + 1]]
        props = _make_props_cantera(species_description)
        species[props["species_name"]] = Species(
            props["molecular_weight"],
            # props['formation_enthalpy'],
            props["total_enthalpy"],
            props["chemical_energy"],
            atoms=props["atoms"],
        )
    return species



from ms_thermo.state import State
from ms_thermo.yk_from_phi import yk_from_phi

__all__ = ["fresh_gas"]


def fresh_gas(temperature, pressure, phi, fuel="KERO"):
    """ Return the the conservative values of the fresh gas from the primitive 
    values.

    :param temperature: the fresh gas temperature
    :type temperature: float
    :param pressure: pressure of the fresh gas
    :type pressure: float
    :param phi: equivalence ratio of the air-fuel mixture
    :type phi: float
    :param fuel: fuel
    :type fuel: string

    :returns:
    	- **rho** - Density
    	- **rhoE** - Conservative energy
    	- **rhoyk** - Dict of conservative mass fractions

    """
    if fuel == "KERO":
        yk = yk_from_phi(phi, 10, 22)
        yk["KERO"] = yk.pop("fuel")

    fresh_gas = State(None, temperature, pressure, yk)
    # print(fresh_gas.keys())
    rho = fresh_gas.rho
    rhoE = fresh_gas.rho * fresh_gas.energy
    rhoyk = dict()
    for specie in fresh_gas._y_k.keys():
        rhoyk[specie] = fresh_gas.rho * fresh_gas._y_k[specie]
    return rho, rhoE, rhoyk
